#!/bin/zsh
# Author: Wolfgang Studer
# Script: POI.sh
# Desc: Prepare data for Passknacker GPI File creation
# Arguments: none to various, see Read.me for possible Options
# Date: 22.03.2021
# Copyright (C) 2021 Wolfgang Studer

# =======================================================
# For sound creation (the convert to MP3) ffmpeg is needed.
# Check out https://www.rickmakes.com/installing-ffmpeg-on-macos-big-sur-on-intel/
# =======================================================

# version of this script
POIshVersion=2021.03.23

# go to Script directory if not already
relativePath="$( dirname $0 )"
cd "$( dirname "$0" )"

# -------------------------------------------------------
# XML Parser function
# -------------------------------------------------------
function read_dom () {
	local ORIGINAL_IFS
	ORIGINAL_IFS=$IFS
	IFS=\>
	read -d \< ENTITY CONTENT
	IFS=$ORIGINAL_IFS
}

# -------------------------------------------------------
# get absolute path of input file
# -------------------------------------------------------
abs_path () {
	echo "$(cd $(dirname "$1");pwd)/$(basename "$1")"
}

# -------------------------------------------------------
# INI Parser function
# https://datacadamia.com/lang/bash/ini
# -------------------------------------------------------
# Get single INI key value
function getIniValue() {
	local IniFile
	local IniValue

	# Init return
	IniValue=""

	# dummy loop
	while true
	do
		# search in user INI
		if [ "${UserIni}" != "" ]
		then
			IniFile="./${UserIni}"
			if [ -f "${IniFile}" ]
			then
				IniValue=$( cat "${IniFile}" | sed '/^#/d' | grep "$1" | cut -d '=' -f 2- )
				IniValue=${${IniValue%${IniValue##*[^[:blank:]]}}#${${IniValue%${IniValue##*[^[:blank:]]}}%%[^[:blank:]]*}}
				if [ "${IniValue}" != "" ]
				then
					break
				fi
			fi
		fi

		# search in factory default INI
		IniFile="./POI.ini"
		if [ -f "${IniFile}" ]
		then
			IniValue=$( cat "${IniFile}" | sed '/^#/d' | grep "$1" | cut -d '=' -f 2- )
			IniValue=${${IniValue%${IniValue##*[^[:blank:]]}}#${${IniValue%${IniValue##*[^[:blank:]]}}%%[^[:blank:]]*}}
			if [ "${IniValue}" != "" ]
			then
				break
			fi
		fi

		# return default
		if [ $# -eq 2 ]
		then
			IniValue="$2"
			break;
		fi

		# end dummy loop
		break
	done

	# return
	echo ${IniValue}
}

# -------------------------------------------------------
# re-Create a folder
# P1 = foldername
# P2 = 1 = delete first, 0 = no delete if already exists
# -------------------------------------------------------
function recreateFolder(){

	# check for parameter
	if [ "$#" -lt 3 ]
	then
		return 1
	fi

	# assign vars
	local RootPath
	local FolderName
	local DeleteFolder
	local Status
	RootPath=$1
	FolderName=$2
	DeleteFolder=$3

	# dummy loop
	Status=0
	while true
	do
		# delete folder if exists
		if [[ -d "${RootPath}/${FolderName}" && ${DeleteFolder} -eq 1 ]]
		then
			rm -rf "${RootPath}/${FolderName}"
			echo "Delete folder: ${RootPath}/${FolderName}" >>"${ProcessLog}"

			# proof deletion
			if [ -d "${RootPath}/${FolderName}" ]
			then
				echo "Folder deletion failed: ${RootPath}/${FolderName}"
				Status=1
				break
			fi
		fi

		# create folder
		if [ ! -d "${RootPath}/${FolderName}" ]
		then
			mkdir "${RootPath}/${FolderName}"
			echo "Create folder: ${RootPath}/${FolderName}" >>"${ProcessLog}"
		fi

		# proof creation
		if [ ! -d "${RootPath}/${FolderName}" ]
		then
			echo "Folder creation failed: ${RootPath}/${FolderName}"
			Status=1
			break
		fi

		# always exit dummy loop
		break
	done

	return ${Status}
}

# -------------------------------------------------------
# find newest file (only 1)
# -------------------------------------------------------
function find_file(){
	local FilePathFound

	FilePathFound=$( find "$1" -name "$2" -print0 | xargs -r -0 ls -1 -tU | head -1 )

	if [ "${FilePathFound}" = "" ]
	then
		return 1
	fi

	echo "${FilePathFound}"
	return 0
}

# -------------------------------------------------------
# check if key exists in array
# -------------------------------------------------------
has_key() {
	local var="${1}[$2]"
	(( ${(P)+${var}} )) && return 0
	return 1
}

# -------------------------------------------------------
# check if variable is numeric
# -------------------------------------------------------
function isnum() {
	# check for numeric values
	awk -v a="$1" 'BEGIN {print (a == a + 0)}';
}

# -------------------------------------------------------
# functions for menu and input parameter handling
# -------------------------------------------------------
function ToggleSwitch() {
	# $1 = current value
	# $2 = new value
	# $3 = 0: new value is absolute, 1: new value is toggle

	# is absolut value or 
	# Toggle Switch Value
	if [ ${3} -eq 0 ]
	then
		return ${2}
	elif [ ${1} -eq 0 ]
	then
		return 1
	else
		return 0
	fi
}

function option_Version() {
	echo ""
	echo "$0, Version: ${POIshVersion}"
	echo ""
	local waitProceed
	read
}

function option_SetTarget() {
	ToggleSwitch ${SetTarget} ${1} ${2}
	SetTarget=${?}
}

function option_AddImages() {
	ToggleSwitch ${AddImages} ${1} ${2}
	AddImages=${?}
}

function option_UseImages() {
	if [ "${AddImages}" -ne 1 ]
	then
		echo ""
		echo "Um diese Funktion nützen zu können muss die Bild Datei Verwendung eingeschalten sein!"
		read
		return 1
	fi

	case "${1}" in
		1)
			if   [ ${UseImages} -eq 0 ];then UseImages=1
			elif [ ${UseImages} -eq 1 ];then UseImages=0
			elif [ ${UseImages} -eq 2 ];then UseImages=3
			elif [ ${UseImages} -eq 3 ];then UseImages=2; fi
			;;
		2)
			if   [ ${UseImages} -eq 0 ];then UseImages=2
			elif [ ${UseImages} -eq 2 ];then UseImages=0
			elif [ ${UseImages} -eq 1 ];then UseImages=3
			elif [ ${UseImages} -eq 3 ];then UseImages=1; fi
			;;
	esac
}

function option_UsePathOnly() {
	if [ "${3}" = "Bild" ] && [ "${AddImages}" -ne 1 ]
	then
		echo ""
		echo "Um diese Funktion nützen zu können muss die Bild Datei Verwendung eingeschalten sein!"
		read
		return 1
	elif [ "${3}" = "Audio" ] && [ "${AddAudio}" -ne 1 ]
	then
		echo ""
		echo "Um diese Funktion nützen zu können muss die Audio Datei Verwendung eingeschalten sein!"
		read
		return 1
	fi

	ToggleSwitch ${UsePathOnly} ${1} ${2}
	UsePathOnly=${?}
}

function option_AddAudio() {
	ToggleSwitch ${AddAudio} ${1} ${2}
	AddAudio=${?}
}

function option_CreateAudio() {
	if [ "${AddAudio}" -ne 1 ]
	then
		echo ""
		echo "Um diese Funktion nützen zu können muss die Audio Datei Verwendung eingeschalten sein!"
		read
		return 1
	fi

	ToggleSwitch ${CreateAudio} ${1} ${2}
	CreateAudio=${?}
}

function option_AudioVoice() {
	if [ "${AddAudio}" -ne 1 ] || [ "${CreateAudio}" -ne 1 ]
	then
		echo ""
		echo "Um diese Funktion nützen zu können muss die Audio Datei Verwendung eingeschalten sein!"
		read
		return 1
	fi

	local choice
	local newSelected
	newSelected=${VoiceName}

	while true
	do
		clear
		echo "Installierte Stimmen:"
		echo ""
		say -v"?"
		echo ""
		echo "Aktuell: ${VoiceName} --> ${newSelected}"
		echo ""
		echo ""
		echo "Deine Stimmen-Wahl"
		echo "oder: W = Weiter"
		echo "      X = Abbruch"
		echo "      T = Test"
		read choice
		case "${choice}" in
			w|W) break ;;
			x|X) return 2 ;;
			t|T) say --voice="${choice}" --rate=${VoiceRate} "Das ist ein Test Wegepunkt" ;;
			*)
				if [ "${choice}" = "" ]
				then
					continue
				fi

				validate_AudioVoice ${choice}
				if [ ${?} -eq 1 ]
				then
					echo -e "${RED}Ungültige Eingabe...${STD}" && sleep 2
					return 1
				fi

				newSelected=${choice}
				;;
		esac
	done

	# warning about Audio Cache
	if [ "${UseAudioCache}" -eq 1 ] && [ "${VoiceName}" -ne "${newSelected}" ]
	then
		clear
		echo ""
		echo "A C H T U N G"
		echo ""
		echo "Durch die Verwendung des Audio Caches wird die Änderung der Stimme"
		echo " nur bei Neu erstellten Audio Dateien verwendet."
		echo "Bestehende Audio Dateien werden aus dem Cache kopiert, ungeachtet der"
		echo " verwendeten Stimme bei deren Erstellung."
		echo ""
		echo "Empfehlung: Audio Cache löschen. Siehe dazu Hilfe Text."
		echo ""
		read
	fi

	# return OK
	VoiceName="${newSelected}"
	return 0
}
function validate_AudioVoice() {
	if [ "$( say -v '?' | grep ${1} )" = "" ]
	then
		return 1
	else
		return 0
	fi
}
function option_AudioRate() {
	if [ "${AddAudio}" -ne 1 ] || [ "${CreateAudio}" -ne 1 ]
	then
		echo ""
		echo "Um diese Funktion nützen zu können muss die Audio Datei Verwendung eingeschalten sein!"
		read
		return 1
	fi

	local choice
	local newSelected
	newSelected=${VoiceRate}

	while true
	do
		clear
		echo "Gewünschte Sprechgeschwindigkeit:"
		echo "( 0 = langsam   / 175 = normal   / 350 = schnell )"
		echo ""
		echo "Aktuell: ${VoiceRate} --> ${newSelected}"
		echo ""
		echo "Deine Wahl"
		echo "oder: W = Weiter"
		echo "      X = Abbruch"
		echo "      T = Test"
		read choice
		case "${choice}" in
			w|W) break ;;
			x|X) return 2 ;;
			t|T) say --voice="${VoiceName}" --rate=${choise} "Das ist ein Test Wegepunkt" ;;
			*)
				if [ "${choice}" = "" ]
				then
					continue
				fi

				if [ ${choice} -lt 1 ] || [ ${choice} -gt 350 ] || [ $( isnum ${choice} ) -ne 1 ] 2>/dev/null
				then
					echo -e "${RED}Ungültige Eingabe...${STD}" && sleep 2
					return 1
				fi

				newSelected=${choice}
				;;
		esac
	done

	# warning about Audio Cache
	if [ "${UseAudioCache}" -eq 1 ] && [ "${VoiceRate}" -ne "${newSelected}" ]
	then
		clear
		echo ""
		echo "A C H T U N G"
		echo ""
		echo "Durch die Verwendung des Audio Caches wird die Änderung der Sprech-"
		echo " geschwindigkeit nur bei Neu erstellten Audio Dateien verwendet."
		echo "Bestehende Audio Dateien werden aus dem Cache kopiert, ungeachtet der"
		echo " eingestellten Sprechgeschwindigkeit bei deren Erstellung."
		echo ""
		echo "Empfehlung: Audio Cache löschen. Siehe dazu Hilfe Text."
		echo ""
		read
	fi

	# return OK
	VoiceRate="${newSelected}"
	return 0
}

function option_UseAudioCache() {
	if [ "${AddAudio}" -ne 1 ]
	then
		echo ""
		echo "Um diese Funktion nützen zu können muss die Audio Datei Verwendung eingeschalten sein!"
		read
		return 1
	fi

	ToggleSwitch ${UseAudioCache} ${1} ${2}
	UseAudioCache=${?}
}

function option_AddIcon() {
	ToggleSwitch ${AddIcon} ${1} ${2}
	AddIcon=${?}
}

function option_ExtendName() {
	ToggleSwitch ${ExtendName} ${1} ${2}
	ExtendName=${?}
}

function option_AddLinks() {
	ToggleSwitch ${AddLinks} ${1} ${2}
	AddLinks=${?}
}

function option_EchoPoi() {
	ToggleSwitch ${EchoPoi} ${1} ${2}
	EchoPoi=${?}
}

function option_CleanTemp() {
	ToggleSwitch ${CleanTemp} ${1} ${2}
	CleanTemp=${?}
}

function option_ShowMenu() {
	ToggleSwitch ${ShowMenu} ${1} ${2}
	ShowMenu=${?}
}

function option_ShowHelp() {
	if [ -f "./Read.me" ]
	then
		clear
		more ./Read.me
	else
	echo ""
		echo "Ups, Die Hilfedatei fehlt. Schau bitte mal nach wo die sein könnte ...."
	echo ""
	fi
}

function option_ListIni() {
	local IniFile

	clear
	echo "Möglichkeiten um dieses Skript mit einem spezifischen INI zu starten:"
	echo ""
	for IniFile in *.ini
	do
		if [ "${IniFile}" != "POI.ini" ] && [ "${IniFile}" != "POI-user.ini" ]
		then
			echo " *)  ${relativePath}/POI.sh --ini=${IniFile}"
		fi
	done
	echo ""
	exit 0
}

function show_menu_item() {
	# $1 = variable holding current setting
	# $2 = flag to steer variable
	# $3 = description of variable
	# $4 = text for switch on
	# $5 = text for switch off

	case "${2}" in
	POI)
		if [ "${1}" -eq 1 ] || [ "${1}" -eq 3 ]
		then
			printf " %-40s |  %-4s  |  %-s\n"  "${3}" " ${2}" "${4}"
		else
			printf " %-40s |  %-4s  |  %-s\n"  "${3}" " ${2}" "${5}"
		fi
		;;
	OWN)
		if [ "${1}" -eq 2 ] || [ "${1}" -eq 3 ]
		then
			printf " %-40s |  %-4s  |  %-s\n"  "${3}" " ${2}" "${4}"
		else
			printf " %-40s |  %-4s  |  %-s\n"  "${3}" " ${2}" "${5}"
		fi
		;;
	AV|AR)
		printf " %-40s |  %-4s  |  %-s\n" "${3}" " ${2}" "${4}"
		;;
	*)
		if [ "$#" -eq 5 ]
		then
			if [ "${1}" -eq 1 ]
			then
				printf " %-40s |  %-4s  |  %-s\n"  "${3}" " ${2}" "${4}"
			else
				printf " %-40s |  %-4s  |  %-s\n"  "${3}" " ${2}" "${5}"
			fi
		else
			printf " %-40s |  %-4s\n" "${3}" " ${2}"

		fi
		;;
	esac
}

function show_menu() {
	clear
	echo "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"
	printf "%-46s %25s\n" " M A I N - M E N U" "Version: ${POIshVersion}"
	echo "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"
	echo ""
	echo "GPI Ziel-Dateien"
	echo "-------------------------------------------------------------------------"
	show_menu_item "${SetTarget}"      "T"   "Zielname des GPI Objektes" "wie GPX" "Fragen"
	show_menu_item "${AddImages}"      "I"   "Bilder verwenden" "√" ""
  if [ "${AddImages}" -eq 1 ]
  then
	show_menu_item "${UseImages}"      "POI" " - POI Bilder verwenden" "√" ""
	show_menu_item "${UseImages}"      "OWN" " - Eigene Bilder verwenden" "√" ""
	show_menu_item "${UsePathOnly}"    "IP"  " > Nur Pfad zu Bilddateien verwenden" "√" ""
  fi

	show_menu_item "${AddAudio}"       "A"   "TourGuide Audio Datei verwenden" "√" ""
  if [ "${AddAudio}" -eq 1 ]
  then
	show_menu_item "${CreateAudio}"    "AC"  " - Audio Datei erstellen" "√" ""
	show_menu_item "${VoiceName}"      "AV"  " - Stimme der Audio Datei" "${VoiceName}"
	show_menu_item "${VoiceRate}"      "AR"  " - Geschwindigkeit (Woerter /Minute)" "${VoiceRate}"
	show_menu_item "${UseAudioCache}"  "CA"  " - TourGuide Audio Cache verwenden" "√" ""
	show_menu_item "${UsePathOnly}"    "AP"  " > Nur Pfad zu Cache-Dateien verwenden" "√" ""
  fi
	show_menu_item "${AddIcon}"        "S"   "Icon Symbol verwenden" "√" ""
	show_menu_item "${ExtendName}"     "N"   "POI Name mit (ID) erweitern" "√" ""
	show_menu_item "${AddLinks}"       "L"   "Link aus Kommentar erstellen" "√" ""

	echo ""
	echo "Aufbereitungsprozess:"
	echo "-------------------------------------------------------------------------"
	show_menu_item "${CleanTemp}"      "ct"  "Temp Daten am Ende bereinigen" "√" "--"
	show_menu_item "${EchoPoi}"        "e"   "Wegpunkte anzeigen" "√" "--"

	echo ""
	echo "Sonstiges:"
	echo "-------------------------------------------------------------------------"
	show_menu_item "1"                 "h"   "Hilfetext anzeigen"
	show_menu_item "1"                 "v"   "Skript Version anzeigen"
	echo ""
	show_menu_item "1"                 "w,W" "Prozess starten"
	show_menu_item "1"                 "x,X" "Beenden"
	echo ""
  if [ "${UserIni}" != "" ]
  then
	echo "Verwendetes INI File: ${UserIni}"
  else
	echo "Verwendetes INI File: POI.ini"
  fi
	echo ">>> ${IniComment}"
	echo ""
}

function read_options(){
	local choice
	echo "Deine Auswahl"
	read choice
	case "${choice}" in
		w|W)
			return 2
			;;
		x|X|q|Q)
			exit 0
			;;
		*)	if [ "${choice}" = $'\e' ]
			then
				# escape key
				exit 0
			elif [ "${choice}" != "" ]
			then
				validate_options ${choice} 1
				if [ ${?} -eq 1 ]
				then
					echo -e "${RED}Ungültige Eingabe...${STD}" && sleep 2
				fi
			fi
			;;
	esac

	# return OK
	return 0
}

function validate_options(){
	# $1 = flag
	# $2 = 0:use absolut value, 1: toggle value
	case $1 in
		-A)  option_AddAudio 0 $2;;
		 A)  option_AddAudio 1 $2;;
		-AC)  option_CreateAudio 0 $2;;
		 AC)  option_CreateAudio 1 $2;;
		 AV | --voice) option_AudioVoice $2;;
		 AR | --rate) option_AudioRate $2;;
		-CA)  option_UseAudioCache 0 $2;;
		 CA)  option_UseAudioCache 1 $2;;
		-AP)  option_UsePathOnly 0 $2 "Audio";;
		 AP)  option_UsePathOnly 1 $2 "Audio";;
		-ct)  option_CleanTemp 0 $2;;
		 ct)  option_CleanTemp 1 $2;;
		-e)  option_EchoPoi 0 $2;;
		 e)  option_EchoPoi 1 $2;;
		-I)  option_AddImages 0 $2;;
		 I)  option_AddImages 1 $2;;
		 OWN) option_UseImages 2 $2;;
		 POI) option_UseImages 1 $2;;
		-IP)  option_UsePathOnly 0 $2 "Bild";;
		 IP)  option_UsePathOnly 1 $2 "Bild";;
		-L)  option_AddLinks 0 $2;;
		 L)  option_AddLinks 1 $2;;
		-m)  option_ShowMenu 0 $2;;
		 m)  option_ShowMenu 1 $2;;
		-N)  option_ExtendName 0 $2;;
		 N)  option_ExtendName 1 $2;;
		-S)  option_AddIcon 0 $2;;
		 S)  option_AddIcon 1 $2;;
		-T)  option_SetTarget 0 $2;;
		 T)  option_SetTarget 1 $2;;

		 listini | -listini | --listini)  option_ListIni "" ;;

		 v | --version) option_Version ;;
		 h | '-?' | '--help') option_ShowHelp ;;
		*)	if [ "${1[1,5]}" != "--ini" ]
			then
				return 1
			fi
			;;
	esac

	# return OK
	return 0
}

function validate_option_values() {
	local STATUS
	STATUS=1

	echo ""

	while true
	do
		if [ ${SetTarget} -ne 0 ] && [ ${SetTarget} -ne 1 ] 2>/dev/null
		then
			echo "Ungültiger Wert für Parameter: Name des GPI Objektes"
			break
		fi
		if [ ${AddImages} -ne 0 ] && [ ${AddImages} -ne 1 ] && [ $( isnum ${AddImages} ) -eq 1 ] 2>/dev/null
		then
			echo "Ungültiger Wert für Parameter: Bilder verwenden"
			break
		else
			if [ ${UseImages} -le 1 ] || [ ${UseImages} -gt 3 ] || [ $( isnum ${AddImages} ) -ne 1 ] 2>/dev/null
			then
				echo "Ungültiger Wert für Parameter: Bilder hinzufügen"
				break
			fi
			if [ ${UsePathOnly} -ne 0 ] && [ ${UsePathOnly} -ne 1 ] 2>/dev/null
			then
				UsePathOnly=0
			fi
		fi
		if [ ${AddAudio} -ne 0 ] && [ ${AddAudio} -ne 1 ] && [ $( isnum ${AddAudio} ) -eq 1 ] 2>/dev/null
		then
			echo "Ungültiger Wert für Parameter: Audio Datei verwenden"
			break
		else
			if [ ${CreateAudio} -ne 0 ] && [ ${CreateAudio} -ne 1 ] && [ $( isnum ${CreateAudio} ) -eq 1 ] 2>/dev/null
			then
				echo "Ungültiger Wert für Parameter: Audio Datei erstellen"
				break
			fi
			# check if ffmpeg is installed
			if [ "${isffmpeg}" -ne 1 ]
			then
				echo "Fehlender MP3 Converter: ffmpeg"
				CreateAudio=0
				break
			fi
			if [ ${CreateAudio} -eq 1 ]
			then
				# validate voice
				validate_AudioVoice ${VoiceName}
				if [ ${?} -eq 1 ]
				then
					VoiceName=""
					echo "Ungültiger Wert für Parameter: Stimme"
					break
				fi
				# validate speech rate
				if [ ${VoiceRate} -lt 1 ] || [ ${VoiceRate} -gt 350 ] || [ $( isnum ${VoiceRate} ) -ne 1 ] 2>/dev/null
				then
					echo "Ungültiger Wert für Parameter: Audio Geschwindigkeit"
					break
				fi
			fi
			# validate audio cache
			if [ ${UseAudioCache} -ne 0 ] && [ ${UseAudioCache} -ne 1 ] 2>/dev/null
			then
				echo "Ungültiger Wert für Parameter: Audio Cache verwenden"
				break
			fi
			# validate use audio path only
			if [ ${UsePathOnly} -ne 0 ] && [ ${UsePathOnly} -ne 1 ] 2>/dev/null
			then
				UsePathOnly=0
			fi
		fi
		if [ ${AddIcon} -ne 0 ] && [ ${AddIcon} -ne 1 ] 2>/dev/null
		then
			echo "Ungültiger Wert für Parameter: Icon hinzufügen"
			break
		fi
		if [ ${ExtendName} -ne 0 ] && [ ${ExtendName} -ne 1 ] 2>/dev/null
		then
			echo "Ungültiger Wert für Parameter: POI Namen mit ID erweitern"
			break
		fi
		if [ ${AddLinks} -ne 0 ] && [ ${AddLinks} -ne 1 ] 2>/dev/null
		then
			AddLinks=0
		fi
		if [ ${EchoPoi} -ne 0 ] && [ ${EchoPoi} -ne 1 ] 2>/dev/null
		then
			EchoPoi=1
		fi
		if [ ${CleanTemp} -ne 0 ] && [ ${CleanTemp} -ne 1 ] 2>/dev/null
		then
			CleanTemp=0
		fi

		# always break dummy loop
		STATUS=0
		break
	done

	# wait for key press to continue
	if [ ${STATUS} -eq 1 ]
	then
		echo ""
		read
	fi
	return ${STATUS}
}

# -------------------------------------------------------
# Initial values to be overwritten by Parameter or menu
# -------------------------------------------------------

# check for dedicated --ini input option
UserIni="POI-user.ini"
if [ $# -gt 0 ]
then
	for par in "$@"
	do
		case "${par[1,5]}" in
		'--ini')
			# as soon as --ini parameter is available, the default user ini file does no longer count
			UserIni=$( echo "${par}" | cut -d"=" -f 2- )
			if [ "${UserIni}" != "-" ] && [ "${UserIni}" != "" ]
			then
				# dedicated ini file set
				break
			fi

			# either no use of user INI wanted or ini file passed does not exists
			UserIni=""
			break
			;;
		esac
	done
fi
# User INI file not found
if [ "${UserIni}" != "" ] && [ ! -f "${UserIni}" ]
then
	UserIni=""
fi

# get defaults option values from ..
# - ini file provided with --ini parameter (if available)
# - POI-user.ini (if available)
# - POI.ini (if available)
# - hardcoded defaults as set below

# set (or not) target GPI name automatic
IniComment=$( getIniValue "comment" "" )

# set (or not) target GPI name automatic
SetTarget=$( getIniValue "SetTarget" 1 )

# use (or not) origin path to images. 0 means use relative path

# add (or not) images
AddImages=$( getIniValue "AddImages" 1 )

# use which images
if [ ${AddImages} -eq 1 ]
then
	UseImages=$( getIniValue "UseImages" 3 )
else
	UseImages=0
fi

# check for ffmpeg, no audio creation if ffmpeg is not available
isffmpeg=0
if [ -x "$(command -v ffmpeg)" ]
then
	isffmpeg=1
fi

# Create (or not) an audio file per POI with spoken Name
AddAudio=$( getIniValue "AddAudio" 0 )
CreateAudio=$( getIniValue "CreateAudio" 0 )
UseAudioCache=0
VoiceName=""
VoiceRate=0
AudioString=""

# set audio related settings
if [ ${AddAudio} -eq 1 ]
then
	UseAudioCache=$( getIniValue "UseAudioCache" 1 )

	if [ ${isffmpeg} -ne 1 ] && [ ${CreateAudio} -eq 1 ]
	then
		echo ""
		echo "FFMPEG als MP3 Audio Converter ist nicht installiert. "
		echo "Eine Audio Datei Erstellung ist ohne ffmpeg nicht möglich."
		echo "Siehe auch im Hilfetext betreffend ffmpeg im Kapitel VORRAUSSETZUNG."
		echo ""

		CreateAudio=0
	fi
	if [ ${CreateAudio} -eq 1 ]
	then
		VoiceName=$( getIniValue "VoiceName" "Anna" )
		VoiceRate=$( getIniValue "VoiceRate" 175 )
		AudioString=$( getIniValue "AudioString" "" )
	elif [ ${UseAudioCache} -ne 1 ]
	then
		# no creation and no cache
		AddAudio=0
	fi
fi

# Use only path to origin image and audio cache
if [ ${AddImages} -eq 1 ] || [ ${AddAudio} -eq 1 ]
then
	UsePathOnly=$( getIniValue "UsePathOnly" 0 )
else
	UsePathOnly=0
fi

# add (or not) Icon
AddIcon=$( getIniValue "AddIcon" 0 )

# Extend (or not) the POI Name with the POI ID >> poiName (poiId)
ExtendName=$( getIniValue "ExtendName" 1 )

# create (or not) links out of comments & description tags
AddLinks=$( getIniValue "AddLinks" 0 )

# Echo (or not) the currently processing POI
EchoPoi=$( getIniValue "EchoPoi" 1 )

# Show the menu
ShowMenu=$( getIniValue "ShowMenu" 1 )

# Remove temp folder at process end
CleanTemp=$( getIniValue "CleanTemp" 1 )

# Build list of link string which should contain numeric POI ID
declare -a linkList
i=0
while true
do
	((i=i+1))
	linkString=$( getIniValue "linkString${i}" "" )
	if [ "${linkString}" = "" ]
	then
		break	# exit loop
	fi

	# add linkString into array
	linkList+="${linkString}"
done

# get the output subfolder names
ImageFolder=$( getIniValue "ImageSubFolder" "images" )

# -------------------------------------------------------
# Parameter handling
# -------------------------------------------------------

# handle parameters when passed
if [ $# -gt 0 ]
then
	for par in "$@"
	do
		case "${par}" in
		-h | '-?' | '--help')	# Help
			if [ $# -eq 1 ]
			then
				option_ShowHelp
			else
				echo "Für Hilfe bitte Script korrekt aufrufen: $0 -h"
			fi
			exit 0
			;;

		*)	# parameter(s) also handled in menu
			validate_options ${par} 0
			if [ ${?} -eq 1 ]
			then
				echo "Ungültiger Parameter: ${par}"
				echo "  > $0 $@"
				exit 1
			fi
			;;
		esac
	done
fi

# -------------------------------------------------------
# Show Menu
# -------------------------------------------------------
if [ ${ShowMenu} -eq 1 ]
then
	while true
	do
		show_menu
		read_options
		if [ ${?} -eq 2 ]
		then
			validate_option_values
			if [ ${?} -eq 0 ]
			then
				break	# start generation process
			fi
		fi
	done
else
	validate_option_values
	if [ ${?} -ne 1 ]
	then
		exit 1
	fi
fi

# -------------------------------------------------------
# Create required files & folders when not available
# -------------------------------------------------------

# start process
StartTS=`date +"%d.%m.%Y  %H:%M:%S"`
echo ""
echo "--------------------------------"
echo "Start: ${StartTS}"
echo "--------------------------------"

# echo errors and standard output also in Log file
FolderName=$( getIniValue "LogFolderName" "" )
if [ "${FolderName}" != "" ]
then
	LogPath="./${FolderName}/"
	recreateFolder "." "${FolderName}" 1
fi
ProcessLog="${LogPath}Process.log"
if [ -f "${ProcessLog}" ]
then
	rm -f "${ProcessLog}"
fi
touch "${ProcessLog}"
########   exec > >(tee ${ProcessLog}) 2>&1

# create log file containing items without images
MissingImagesLog="${LogPath}MissingImages.log"
if [ -f "${MissingImagesLog}" ]
then
	rm -f "${MissingImagesLog}"
fi
touch "${MissingImagesLog}"

# define temp folder
FolderName=$( getIniValue "TempFolderName" "temp" )
TempPath="./${FolderName}/"
recreateFolder  "." "${FolderName}" 1

# define result output folder
FolderName=$( getIniValue "OutputFolderName" "output" )
OutputPath="./${FolderName}/"
recreateFolder  "." "${FolderName}" 1

# define input folder with GPX files (*.gpx)
FolderName=$( getIniValue "InputFolderName" "INPUT" )
InputPath="./${FolderName}/"
recreateFolder  "." "${FolderName}" 0

# define folder with origin Passknacker images (${id}.jpg)
FolderName=$( getIniValue "ImageFolderNamePOI" "IMAGES.POI" )
ImageFolderPOI="./${FolderName}/"
recreateFolder  "." "${FolderName}" 0

# define folder with user specific images (${id}.jpg), overwrites POI origin imges in result
FolderName=$( getIniValue "ImageFolderNameOWN" "IMAGES.OWN" )
ImageFolderOWN="./${FolderName}/"
recreateFolder  "." "${FolderName}" 0

# define folder with bmp symbol icons
FolderName=$( getIniValue "IconFolderName" "ICONS" )
IconPath="./${FolderName}/"
recreateFolder  "." "${FolderName}" 0

# define folder for MP3 Audio TourGuide files cache
if [ ${UseAudioCache} -eq 1 ]
then
	FolderName=$( getIniValue "CacheFolderName" "CACHE" )
	CachePath="./${FolderName}/"
	recreateFolder  "." "${FolderName}" 0
fi

# define CSV file containing all (road/xxx, driven/undriven) POI's
# used for issue logging
FolderName=$( getIniValue "CsvFolderName" "CSV" )
CsvPath="./${FolderName}/"
recreateFolder  "." "${FolderName}" 0

CsvFile=""
if [ -d "${CsvPath}" ]
then
	FileName=$( getIniValue "CsvFileName" "POI.csv" )
	CsvFile="${CsvPath}${FileName}"
	if [ -f "${CsvFile}" ]
	then
		# check for UTF8 format
		if [ ! $(file "${CsvFile}" | grep "UTF-")]
		then
			# convert to UTF-8 if not already in this format
			iconv -f ISO-8859-1 -t UTF-8//TRANSLIT "${CsvFile}" > "${TempPath}/UTF8_${FileName}"
			CsvFile="${TempPath}/UTF8_${FileName}"
		fi
	fi
	if [ -f "${CsvFile}" ]
	then
		# replace ' by `
		sed -i '' -e "s|\'|\`|g" "${CsvFile}"
	else
		CsvFile=""
	fi
fi

# -------------------------------------------------------
# Ask for target GPI file name => Name of folder, GPX and BMP
# -------------------------------------------------------
# Build list of GPI names
typeset -A gpiNameList
typeset -a gpiIdList
for InputFile in $( ls "${InputPath}" )
do
	if [ -f "${InputPath}${InputFile}" ] && [ "${(U)InputFile[-3,-1]}" = "GPX" ]
	then
		# get FileName without extension
		FileName=$(basename "${InputFile:0:-5}")

		# add filename to naming array
		gpiNameList[${FileName}]="${FileName}"
		gpiIdList+="${FileName}"

	elif [ -d "${InputPath}${InputFile}" ]
	then
		# the folder must exists at least 1 GPX file
		for ClusterFile in $(ls "${InputPath}${InputFile}" )
		do
			if [ -f "${InputPath}${InputFile}/${ClusterFile}" ] && [ "${(U)ClusterFile[-3,-1]}" = "GPX" ]
			then
				# add foldername to naming array
				gpiNameList[${InputFile}]="${InputFile}"
				gpiIdList+="${InputFile}"

				# end of file loop
				break
			fi
		done
	fi
done


# no files found
if [ ${i} -eq 0 ]
then
	echo ""
	echo "Keine GPX Dateien zur Verarbeitung gefunden."
	echo "Programmausführung wird abgebrochen."
	echo ""
	exit 0
fi


if [ "${SetTarget}" -eq 0 ]
then
	while true
	do
		clear
		echo ""
		echo " Bitte passe den gewünschten Namen jedes GPI Objektes an:"
		echo " ========================================================"
		echo ""
		printf " %2s: %-30s = %s \n" "#" "GPX Datei Name" "GPI Objekt Name"
		echo " --------------------------------------------------------"
		i=0
		for fileName in ${gpiIdList}
		do
			((i=i+1))
			printf " %2s: %-30s = %s \n" "${i}" "${fileName}" "${gpiNameList[${fileName}]}"
		done

		echo ""
		echo " Bitte Nummer (+Return) der GPX Datei angeben um GPI Namen zu ändern"
		echo " oder 'w' für weiter:"
		read choice
		case "${choice}" in
			w|W) break ;;
			*)	if [ "${choice}" != ""]
				then
					has_key gpiIdList choice
					if [ ${?} -eq 1 ]
					then
						echo ""
						echo "Ein Listeintrag mit der Nummer ${choice} existiert nicht."
						echo "Bitte erneut versuchen (Weiter mit Return)"
						read
						continue
					fi

					gpxName="${gpiIdList[${choice}]}"
					echo ""
					echo "GPX = ${gpxName}"
					echo "GPI = ${gpiNameList[${gpxName}]}"
					echo "Bestätigen ohne Eingabe behält den aktuellen Namen"
					read choice
					if [ "${choice}" != "" ]
					then
						# save new name
						gpiNameList[${gpxName}]="${choice}"
					fi
				fi
		esac
	done
fi
unset gpiIdList

# -------------------------------------------------------
# Data processing
# -------------------------------------------------------

# Add "TourGuide" when Audio to Folder, GPX, BMP, IMG, MP3
TourGuideString=""
if [ ${AddAudio} -eq 1 ] && [ ${UsePathOnly} -eq 0 ]
then
	TourGuideString=$( getIniValue "TourGuideString" "" )
fi

# Show options in log
	echo ""
	echo "--------------------------------"
	echo "Verwendete Optionen"
	echo "--------------------------------"
	echo "SetTarget      : ${SetTarget}"
	echo "AddImages      : ${AddImages}"
	echo "UseImages      : ${UseImages}"
	echo "UsePathOnly   : ${UsePathOnly}"
	echo "AddAudio       : ${AddAudio}"
if [ ${CreateAudio} -eq 1 ]
then
	echo "CreateAudio    : ${CreateAudio}"
	echo "VoiceName      : ${VoiceName}"
	echo "VoiceRate      : ${VoiceRate}"
	echo "AudioString    : ${AudioString}"
	echo "UseAudioCache  : ${UseAudioCache}"
	echo "UsePathOnly   : ${UsePathOnly}"
	echo "TourGuideString: ${TourGuideString}"
fi
	echo "AddIcon        : ${AddIcon}"
	echo "ExtendName     : ${ExtendName}"
	echo "AddLinks       : ${AddLinks}"
	echo "EchoPoi        : ${EchoPoi}"
	echo "ShowMenu       : ${ShowMenu}"
	echo "CleanTemp      : ${CleanTemp}"
	echo "Comment        : ${IniComment}"

# copy all INPUT GPX files into temp folder
echo ""
echo "--------------------------------"
printf "%s: %s \n" "$(date +'%d.%m.%Y  %H:%M:%S')" "Bereite GPX Datei(en) vor (Kann dauern, bitte Geduld haben)"
echo "--------------------------------"

# loop all files & clusters
for InputFile in $( ls "${InputPath}" )
do
	if [ -f "${InputPath}${InputFile}" ] && [ "${(U)InputFile[-3,-1]}" = "GPX" ]
	then
		# get FileName without extension
		FileName=$(basename "${InputFile:0:-5}")

		# Copy Input file to temp
		echo ""
		printf "%s: %s: %s\n" "$(date +'%d.%m.%Y  %H:%M:%S')" "GPX Datei" "${gpxName}.gpx"
		cp "${InputPath}${InputFile}" "${TempPath}${gpxName}.gpx"

		# adjust all ' to `
		printf "%s: %s \n" "$(date +'%d.%m.%Y  %H:%M:%S')" "Korrekter Apostroph setzen."
		sed -i '' -e "s|\'|’|g" "${TempPath}${gpxName}.gpx"

	elif [ -d "${InputPath}${InputFile}" ]
	then
		# loop files in cluster folder and merge them together
		echo ""
		printf "%s: %s: %s\n" "$(date +'%d.%m.%Y  %H:%M:%S')" "GPX Cluster" "${InputFile}/*.gpx"
		i=0

		for ClusterFile in $(ls "${InputPath}${InputFile}" )
		do
			if [ -f "${InputPath}${InputFile}/${ClusterFile}" ] && [ "${(U)ClusterFile[-3,-1]}" = "GPX" ]
			then

				# leave first file as is
				((i=i+1))
				if [ ${i} -gt 1 ]
				then
					# copy file as temp file
					cp "${InputPath}${InputFile}/${ClusterFile}" "${TempPath}TEMP.gpx"

					# remove heading lines in temp file
					substring="<wpt "
					while IFS='' read -r LINE || [ -n "${LINE}" ]
					do
						# stop loop when reaching first WPT tag
						if test "${LINE#*$substring}" != "$LINE"
						then
							break
						fi

						# delete line
						sed -i '' -e '1d' "${TempPath}TEMP.gpx"
					done < "${InputPath}${InputFile}/${ClusterFile}"

					# merge file
					cat "${TempPath}TEMP.gpx" >> "${TempPath}${InputFile}.gpx"

					# delete temp gpx file
					rm -f "${TempPath}TEMP.gpx"
				else
					# first file
					cp "${InputPath}${InputFile}/${ClusterFile}" "${TempPath}${InputFile}.gpx"
				fi
			fi
		done

		if [ -f "${TempPath}${InputFile}.gpx" ]
		then
			# adjust all ' to `
			printf "%s: %s \n" "$(date +'%d.%m.%Y  %H:%M:%S')" " - Korrekter Apostroph setzen."
			sed -i '' -e "s/\'/’/g" "${TempPath}${InputFile}.gpx"

			# remove all end of files tags but add a final one
			printf "%s: %s \n" "$(date +'%d.%m.%Y  %H:%M:%S')" " - Abschluss-Tags bereinigen."
			sed -i '' -e "/<\/gpx>/d" "${TempPath}${InputFile}.gpx"
			echo "</gpx>" >> "${TempPath}${InputFile}.gpx"
		fi
	fi
done

# end
echo ""
printf "%s: %s \n" "$(date +'%d.%m.%Y  %H:%M:%S')" "GPX Datei(en) Vorbereitung beendet"


# loop over available GPX files and process one by one
for InputFile in ${TempPath}*.gpx
do
	# get FileName without extension
	gpxName=$(basename -s .gpx ${InputFile})
	gpiName="${gpiNameList[${gpxName}]}"

	# Start
	if [ ${EchoPoi} -eq 1 ]
	then
		echo ""
		echo "--------------------------------"
		printf "%s: %s: %s \n" "$(date +'%d.%m.%Y  %H:%M:%S')" "Starte Bearbeitung" "${InputFile}"
		echo "                      ${gpxName} ==> ${TourGuideString}${gpiName}"
		echo "--------------------------------"
	fi

	# Basic validation: are name and link tags available ?
	if [ "$(grep '<name>' ${InputFile})" = "" ] || [ "$(grep '<link ' ${InputFile})" = "" ]
	then
		MissingTag="<link...>"
		if [ "$(grep '<name>' ${InputFile})" = "" ]; then MissingTag="<name>"; fi

		echo ""
		echo "Die Datei
	'${InputFile}'
kann nicht verarbeitet werden, da Ihr das ${MissingTag} tag fehlt."
		echo ""
		echo ""
		echo "Bitte mit der Enter Taste bestätigen!"
		read
		continue
	fi

	# create output target folder for current GPX (and additional files) if not yet exists
	DataPath="${OutputPath}${TourGuideString}${gpiName}"
	recreateFolder "${OutputPath}" "${TourGuideString}${gpiName}" 1

	# copy also icon if available
	if [ ${AddIcon} -eq 1 ]
	then
		if [ -f "${IconPath}${gpxName}.bmp" ]
		then
			cp "${IconPath}${gpxName}.bmp" "${DataPath}/${TourGuideString}${gpiName}.bmp"
		elif [ "${gpxName}" != "${gpiName}" ] && [ -f "${IconPath}${gpixName}.bmp" ]
		then
			cp "${IconPath}${gpiName}.bmp" "${DataPath}/${TourGuideString}${gpiName}.bmp"
		else
			echo ""
			echo "Icon fehlt"
			echo "----------"
			if [ "${gpxName}" = "${gpiName}" ]
			then
				echo "${IconPath}${gpxName}.bmp"
			else
				echo "${IconPath}${gpxName}.bmp oder ${gpiName}.bmp"
			fi
			echo ""
		fi
	fi

	# copy current processing file into GPX folder
	printf "%s: %s \n" "$(date +'%d.%m.%Y  %H:%M:%S')" "GPX Datei kopieren"
	GpxFile="${DataPath}/${TourGuideString}${gpiName}.gpx"
	cp "${InputFile}" "${GpxFile}"

	# collect URL's inside GPX
	if [ ${AddLinks} -eq 1 ]
	then
		grep -io 'http://[a-z.-]*/[a-z0-9+-_%/]*' "${InputFile}"  >  "${TempPath}urls.raw"
		grep -io 'https://[a-z.-]*/[a-z0-9+-_%/]*' "${InputFile}" >> "${TempPath}urls.raw"
		grep -io 'www.[a-z.-]*/[a-z0-9+-_%/]*' "${InputFile}"     >> "${TempPath}urls.raw"

		# cleanup url list
		sed -i '' -e 's|</cmt>||g' -e 's/^[ \t]*//;s/[ \t]*$//' -e 's|http://||g' "${TempPath}urls.raw"
		sed -i '' -e '/passknacker.com/d' -e '/garmin.com/d'  -e '/topografix/d' "${TempPath}urls.raw"
		sort "${TempPath}urls.raw" | uniq > "${TempPath}urls.sort"

		# add urls to array
		typeset -A urlsList
		while IFS='' read -r LINE || [ -n "${LINE}" ]
		do
			# save key without https:// is available, but keep val as is
			has_key urlsList ${LINE/https:\/\//}
			if [ ${?} -eq 1 ]
			then
				# not yet in list
				urlsList[${LINE/https:\/\//}]="${LINE}"
			fi
		done < "${TempPath}urls.sort"
	fi

	# create sub folders for images when not TourGuide
	if [ ${AddImages} -eq 1 ] && [ ${AddAudio} -eq 0 ] && [ ${UseImagesPath} -eq 0 ]
	then
		recreateFolder "${DataPath}" "${ImageFolder}" 1
		DataTargetFolder="${ImageFolder}"
	fi

	# init flag for missing images
	missingImages=0

	# loop name file to identify POI ID's
	wptStart=""
	wptEndReached=0
	HasAIFF=0

	while read_dom
	do
		case $(echo ${ENTITY} | xargs) in
		"/gpx")
			# end of file reached
			break
			;;

		"name")
			# save waypoint name, replace ' by `
			wptName="${CONTENT}"
			;;

		"cmt"|"desc")
			if [ ${AddLinks} -eq 1 ]
			then
				# search urls in cmt and desc tags and add as external links
				for key value in ${(kv)urlsList}
				do
					# check if current url is in content
					if test "${CONTENT#*$key}" != "$CONTENT"
					then
						# add to an array all links found
						# add as external link is done later as wptID not yet known
						urlsListWpt+=${value}
					fi
				done
			fi
			;;

		"/wpt")
			# waypoint end reached, process current waypoint data
			if [ "${wptName}" = "" ]
			then
				# no name found
				echo "Name fehlt: [${wptId}] = ${wptStart}"
				continue
			fi
			if [ ${wptId} -lt 1 ] || [ ${wptId} -gt 99999 ] || [ $( isnum ${wptId} ) -ne 1 ] 2>/dev/null
			then
				err=0
				if [ ${wptId} -lt 1 ];then err="Keine ID gefunden"
				elif [ ${wptId} -gt 99999 ];then err="ID grösser als erlaubt"
				elif [ $( isnum ${wptId} ) -ne 1 ];then err="ID ist keine Nummer"
				else err="Sonstiger ID Fehler"; fi

				# not a number
				echo "Ungültige Id: [${wptId}] = ${wptName} | ${err}"
				continue
			fi

			# show currently processed waypoint
			if [ ${EchoPoi} -eq 1 ]
			then
				printf " * %5d : %s\n" "${wptId}" "${wptName}"
			fi

			# add wptId to name
			if [ ${ExtendName} -eq 1 ]
			then
				sed -i '' -e 's|<name>'"${wptName}"'</name>|<name>'"${wptName}"' - '"${wptId}"'</name>|' "${GpxFile}"
			fi

			# create TourGuide target folder for images & audio
			if [ ${CreateAudio} -eq 1 ] && [ ${UsePathOnly} -eq 0 ]
			then
				recreateFolder "${DataPath}" "${TourGuideString}${wptId}" 1
				DataTargetFolder="${TourGuideString}${wptId}"
			fi

			# handle images
			if [ ${AddImages} -eq 1 ]
			then
				# find images in data pool
				ImageFile=""
				while [ "${ImageFile}" = "" ]
				do
					if [ ${UseImages} -eq 2 ] || [ ${UseImages} -eq 3 ]
					then
						ImageFile=$( find_file "${ImageFolderOWN}" "${wptId}.[jJ][pP][gG]" )
						if [ ${?} -eq 0 ]
						then
							break	# image found
						fi
					fi
					if [ ${UseImages} -eq 1 ] || [ ${UseImages} -eq 3 ]
					then
						ImageFile=$( find_file "${ImageFolderPOI}" "${wptId}.[jJ][pP][gG]" )
						if [ ${?} -eq 0 ]
						then
							break	# image found
						fi
					fi

					# break anyway
					ImageFile=""
					break
				done

				if [ ${UsePathOnly} -eq 0 ] && [ "${ImageFile}" != "" ]
				then
					# copy image found
					cp -f "${ImageFile}" "${DataPath}/${DataTargetFolder}/${TourGuideString}${wptId}.jpg"

					# add link to image as relative path
					sed -i '' -e 's|<link img='${wptId}'>|<link href="'${DataTargetFolder}'/'${TourGuideString}${wptId}'.jpg\"/>|' "${GpxFile}"

				elif [ ${UsePathOnly} -eq 0 ] && [ "${ImageFile}" = "" ]
				then
					# report missing image
					if [ ${missingImages} -eq 0 ]
					then
						missingImages=1
						echo ""                                         >> "${MissingImagesLog}"
						echo "POI Bild fehlt (${gpiName})"              >> "${MissingImagesLog}"
						echo "----------------------------------------" >> "${MissingImagesLog}"
					  if [ -f "${CsvFile}" ]
					  then
						echo "ID;Name;POI CSV"                          >> "${MissingImagesLog}"
					  else
						echo "ID;Name"                                  >> "${MissingImagesLog}"
					  fi
					fi

					if [ -f "${CsvFile}" ]
					then
						c=$(grep "${wptName}" ${CsvFile})
						echo "$wptId;${wptName};${c}" >> "${MissingImagesLog}"
					else
						echo "$wptId;${wptName}"      >> "${MissingImagesLog}"
					fi

				elif [ ${UsePathOnly} -eq 1 ] && [ "${ImageFile}" != "" ]
				then
					# add link to origin image
					sed -i '' -e 's|<link img='${wptId}'>|<link href="'$(abs_path ${ImageFile})'\"/>|' "${GpxFile}"
				fi
			fi

			# handle audio
			if [ ${AddAudio} -eq 1 ]
			then
				if [ ${UsePathOnly} -eq 0 ] && [ ${CreateAudio} -eq 1 ]
				then
					# audio file already available in cache
					if [ ${UseAudioCache} -eq 1 ] && [ -f "${CachePath}/${wptId}.mp3" ]
					then
						# copy MP3 file from cache
						cp "${CachePath}/${wptId}.mp3" "${DataPath}/${DataTargetFolder}/${TourGuideString}${wptId}.mp3"

						# add link to audio file
						sed -i '' -e 's|<link mp3='${wptId}'>|<link href="'${TourGuideString}${wptId}'/'${TourGuideString}${wptId}'.mp3\"/>|' "${GpxFile}"

					else
						# create sound file as mp3 (say does not support mp3 output directly)

						# replace special chars characters inside names, then trim leading/trailing spaces
						AudioName=$(echo ${wptName} | sed -f ./POI_AUDIO.sed | xargs)

						# create a temp audio file
						say --voice="${VoiceName}" --rate=${VoiceRate} "${AudioString} ${AudioName}" --output-file="${DataPath}/${DataTargetFolder}/${wptId}.aiff"

						# mp3 conversion is done later with ffmpeg,
						# because ffmpeg destroys read_dom function loop when done here in ...
						# link to audio file is done at time of mp3 file creation
						HasAIFF=1
					fi

				elif [ ${UsePathOnly} -eq 1 ] && [ ${UseAudioCache} -eq 1 ]
				then
					# audio file already available in cache
					if [ ${UseAudioCache} -eq 1 ] && [ -f "${CachePath}/${wptId}.mp3" ]
					then
						# add link to audio file in cache
						sed -i '' -e 's|<link mp3='${wptId}'>|<link href="'$(abs_path ${CachePath}/${wptId}'.mp3')'\"/>|' "${GpxFile}"

					else
						# create sound file as mp3 (say does not support mp3 output directly)

						# replace special chars characters inside names, then trim leading/trailing spaces
						AudioName=$(echo ${wptName} | sed -f ./POI_AUDIO.sed | xargs)

						# create a temp audio file directly into cache
						say --voice="${VoiceName}" --rate=${VoiceRate} "${AudioString} ${AudioName}" --output-file="${CachePath}/${wptId}.aiff"

						# mp3 conversion is done later with ffmpeg,
						# because ffmpeg destroys read_dom function loop when done here in ...
						# link to audio file is done at time of mp3 file creation
						HasAIFF=1
					fi
				fi
			fi

			# Add links inside waypoint comment or description as external links
			if [ ${AddLinks} -eq 1 ]
			then
				for value in ${urlsListWpt[@]}
				do
					# add as external link
					sed -i '' -e 's|<link ext='${wptId}'>|<link href="'${value}'"/>#n#<link ext='${wptId}'>|' "${GpxFile}"
				done
			fi
			unset urlsListWpt
			;;
		*)
			# waypoint start
			if [ "${ENTITY[1,4]}" = "wpt " ]
			then
				wptStart="${ENTITY}"
				wptName=""
				wptId=0
				typeset -a urlsListWpt

				continue
			fi

			# ignore all not link tags
			if [ "${ENTITY[1,5]}" != "link " ]
			then
				continue
			fi

			# get numeric id from valid link tags
			# <link href="https://passknacker.com/paesse_details.php?pass=1"/>
			# <link href="images\1.jpg"/>
			# <link href="images/1.jpg"/>
			for vl in "${linkList[@]}"
			do
				# valid link tag not in current link
				if [[ "${ENTITY}" != *"${vl}"* ]]
				then
					continue
				fi

				# get id
				wptId=$(grep -oE "[[:digit:]]{1,}" <<< "${ENTITY}" )

				# remove dummy links when already exists
				sed -i '' -e '/<link img='${wptId}'>/d' -e '/<link mp3='${wptId}'>/d' "${GpxFile}"

				# add dummy links, #n# will be replaced later by a linefeed
				if [ "${CreateAudio}" -eq 1 ] || [ "${AddImages}" -eq 1 ] || [ ${UseOriginPath} -eq 1 ]
				then
					sed -i '' -e 's|<'${ENTITY}'>|<'${ENTITY}'>#n#<link img='${wptId}'>#n#<link mp3='${wptId}'>#n#<link ext='${wptId}'>|' "${GpxFile}"
				fi

				if [ ${wptId} -gt 0 ] && [ $( isnum ${wptId} ) -eq 1 ] 2>/dev/null
				then
					break	# exit loop linkList, read next ENTITY
				fi
			done
			;;
		esac
	done < "${InputFile}"


	# convert temp created audio file as TourGuide MP3
	# This process has to be done at end as otherwise the read_dom loop crashes
	if [ ${HasAIFF} -eq 1 ]
	then
		echo ""
		echo "--------------------------------"
		printf "%s: %s \n" "$(date +'%d.%m.%Y  %H:%M:%S')" "Erstelle MP3 TourGuide Audio Dateien"
		echo "--------------------------------"

		if [ ${UsePathOnly} -eq 0 ]
		then
			for AiffFile in ${DataPath}/**/*.aiff
			do
				[ -e "${AiffFile}" ] || continue
				# get AudioName without extension
				wptId=$(basename -s .aiff ${AiffFile})
				FolderName=${AiffFile%/*}
				# convert AIFF into MP3
				echo "Erstelle: ${FolderName} / ${TourGuideString}${wptId}.mp3"
				ffmpeg -loglevel +error -i "${AiffFile}" "${FolderName}/${TourGuideString}${wptId}.mp3"
				#remove old aiff file
				rm -f "${AiffFile}"
				# add new created mp3 audio file into cache
				if [ "${UseAudioCache}" -eq 1 ]
				then
					cp "${FolderName}/${TourGuideString}${wptId}.mp3" "${CachePath}/${wptId}.mp3"
				fi

				# add link to audio file
				if [ -f "${FolderName}/${TourGuideString}${wptId}.mp3" ]
				then
					if [ ${UseOriginPath} -eq 1 ] && [ "${UseAudioCache}" -eq 1 ]
					then
						sed -i '' -e 's|<link mp3='${wptId}'>|<link href="'$(abs_path ${CachePath}/${wptId}'.mp3')'\"/>|' "${GpxFile}"
					else
						sed -i '' -e 's|<link mp3='${wptId}'>|<link href="'${TourGuideString}${wptId}'/'${TourGuideString}${wptId}'.mp3\"/>|' "${GpxFile}"
					fi
				fi
			done

		else
			for AiffFile in ${CachePath}/*.aiff
			do
				[ -e "${AiffFile}" ] || continue
				# get AudioName without extension
				wptId=$(basename -s .aiff ${AiffFile})
				# convert AIFF into MP3
				echo "Erstelle: CACHE / ${wptId}.mp3"
				ffmpeg -loglevel +error -i "${AiffFile}" "${CachePath}/${wptId}.mp3"
				#remove old aiff file
				rm -f "${AiffFile}"

				# add link to audio file
				sed -i '' -e 's|<link mp3='${wptId}'>|<link href="'$(abs_path ${CachePath}/${wptId}'.mp3')'\"/>|' "${GpxFile}"
			done
		fi

		echo ""
		printf "%s: %s \n" "$(date +'%d.%m.%Y  %H:%M:%S')" "Erstellung MP3 TourGuide Audio Dateien beendet"
	fi

	# cleanup GPX
	echo ""
	echo "--------------------------------"
	printf "%s: %s \n" "$(date +'%d.%m.%Y  %H:%M:%S')" "Bereinige Ziel GPX Datei (Kann dauern, bitte Geduld haben)"
	echo "--------------------------------"

	# Add Linefeeds
	printf "%s: %s \n" "$(date +'%d.%m.%Y  %H:%M:%S')" " - Zeilenumbrüche einfügen."
	sed -i '' -e 's|#n#|\n    |g' "${GpxFile}"

	# File list of possible GPX correction files to be executed - if available - at file loop end
	typeset -a sedFileList
	sedFileList+="${InputPath}${gpiName}/${gpiName}.sed"
	sedFileList+="${InputPath}${gpiName}/POI.sed"
	sedFileList+="${InputPath}${gpiName}.sed"
	sedFileList+="${InputPath}${gpxName}.sed"
	sedFileList+="${InputPath}POI.sed"
	sedFileList+="./${gpiName}.sed"
	sedFileList+="./POI.sed"

	# run GPX corrections if exists
	for fileName in ${sedFileList}
	do
		if [ -f "${fileName}" ]
		then
			curPath=
			printf "%s: %s (%s)\n" "$(date +'%d.%m.%Y  %H:%M:%S')" " - GPX Korrekturen." "${fileName:s/$curPath/.}"
			sed -i '' -f "${fileName}" "${GpxFile}"
		fi
	done

	unset sedFileList

	# remove dead links
	printf "%s: %s \n" "$(date +'%d.%m.%Y  %H:%M:%S')" " - Tote Links löschen."
	sed -i '' -e '/<link img=/d' -e '/<link mp3=/d' -e '/<link ext=/d' "${GpxFile}"

	# end
	printf "%s: %s \n" "$(date +'%d.%m.%Y  %H:%M:%S')" "GPX Datei Bereinigung beendet"

# process next input file
done

	echo ""
	echo "--------------------------------"

# cleanup temp folder
	if [ ${CleanTemp} -eq 1 ]
	then
		rm -rf "${TempPath}"
		echo ""
		printf "%s: %s: %s \n" "$(date +'%d.%m.%Y  %H:%M:%S')" "Lösche Verzeichnis" "${TempPath}"
		echo ""
	fi

# end process
	if [ ${missingImages} -eq 1 ]
	then
		cat "${MissingImagesLog}"
	fi

	echo ""
	echo "--------------------------------"
	echo "POI GPI Aufbereitungs-Prozess beendet"
	echo ""
	echo "Start: ${StartTS}"
	echo "Ende : $(date +'%d.%m.%Y  %H:%M:%S')"
	echo ""
	echo "Die aufbereiteten Verzeichnisse für den Garmin PoiLoader sind hier:"
	echo "    $(abs_path ${OutputPath})"
	echo "--------------------------------"


exit 0
